//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: Sat Jan 19 10:38:26 PST 2002
// Last Modified: Sat Jan 19 10:38:29 PST 2002
// Filename:      SoundFileWrite.cpp
// Based On:      http://sig.sapp.org/src/sigSignal/SoundFileOut.cpp
// Syntax:        C++ 
// Copyright:     Copyright 2002 Craig Stuart Sapp
//

#include "SoundFileWrite.h"
#include <iostream.h>
#include <stdlib.h>
#include <string.h>



//////////////////////////////
//
// SoundFileWrite::SoundFileWrite --
//

SoundFileWrite::SoundFileWrite(const char* aFileName, SoundHeader& header) {
   filename = new char[strlen(aFileName)+1];
   strcpy(filename, aFileName);
   outSample = 0;
   #ifdef VISUAL            /* for stupid LF-CR prevention in Windows */
      outFile.open(filename, ios::out | ios::noreplace | ios::binary);
   #else
      outFile.open(filename, ios::out | ios::noreplace);
   #endif
   if (!outFile.is_open()) {
      cerr << "Error: cannot write soundfile " << filename 
           << ". Perhaps the file already exists?" << endl;
      exit(1);
   }
   sndHeader = header;
   sndHeader.modifyFormat(aFileName);
   outFile << sndHeader;
   outFile.seekp(sndHeader.getDataByteOffset(), ios::beg);
   sampleCount = 0;
   determineWritingFunction();
}



//////////////////////////////
//
// SoundFileWrite::~SoundFileWrite -- close the sound file.
//


SoundFileWrite::~SoundFileWrite() {
   close();
}



//////////////////////////////
//
// SoundFileWrite::close -- write the size of the file in the header
//     and close the sound file.
//

void SoundFileWrite::close(void) {
   if (sndHeader.getSamples() != sampleCount/sndHeader.getChannels()) {
      switch (sndHeader.getOutputType()) {
         case TYPE_SND:
            // write the number of bytes in the data:
            outFile.seekp(8, ios::beg);   
            outFile.writeBigEndian((long)
               (sampleCount/sndHeader.getChannels()*sndHeader.getBlockAlign()));
            break;
         case TYPE_WAV_PCM:
            outFile.seekp(8, ios::beg);
            // write the number of bytes in the file minus a few:
            outFile.seekp(4, ios::beg);   
            outFile.writeLittleEndian((long)
               (36 + sampleCount/sndHeader.getChannels() * 
                  sndHeader.getBlockAlign()));
            // write the number of sound data bytes:
            outFile.seekp(40, ios::beg);
            outFile.writeLittleEndian((long)
               (sampleCount/sndHeader.getChannels()*sndHeader.getBlockAlign()));
            break;
         default:
           cerr << "Unknown output soundfile type: "
                << sndHeader.getOutputType() << endl;
           exit(1);
      }
   }
   outFile.close();
   sampleCount = 0;
}



//////////////////////////////
//
// SoundFileWrite::getFilename -- will never return a NULL pointer.
//

const char* SoundFileWrite::getFilename(void) {
   if (filename == NULL) {
      return "";
   } else {
      return filename;
   }
}



//////////////////////////////
//
// SoundFileWrite::writeSample16Bit --
//

void SoundFileWrite::writeSample16Bit(short sample) {
   switch (endianType) {
      case SIG_BIG_ENDIAN:    outFile.writeBigEndian(sample);    break;
      case SIG_LITTLE_ENDIAN: outFile.writeLittleEndian(sample); break;
   }
   sampleCount++;
}
     


//////////////////////////////
//
// SoundFileWrite::writeSampleDouble -- convert to a 16-bit value and
//    write to file.
//

void SoundFileWrite::writeSampleDouble(double sample) {
   writeSample(outFile, sample);
   sampleCount++;
}
     


///////////////////////////////////////////////////////////////////////////
//
// private functions:
//

uchar linear16ToMulaw8(short aValue);

// big endian samples
void  soundWriteSample8M_B      (FileIO& outFile, double aSample);
void  soundWriteSample8L_B      (FileIO& outFile, double aSample);
void  soundWriteSample16L_B     (FileIO& outFile, double aSample);
void  soundWriteSample24L_B     (FileIO& outFile, double aSample);
void  soundWriteSample32L_B     (FileIO& outFile, double aSample);
void  soundWriteSample32F_B     (FileIO& outFile, double aSample);
void  soundWriteSample64F_B     (FileIO& outFile, double aSample);

// little endian samples
void  soundWriteSample8M_L      (FileIO& outFile, double aSample);
void  soundWriteSample8L_L      (FileIO& outFile, double aSample);
void  soundWriteSample16L_L     (FileIO& outFile, double aSample);
void  soundWriteSample24L_L     (FileIO& outFile, double aSample);
void  soundWriteSample32L_L     (FileIO& outFile, double aSample);
void  soundWriteSample32F_L     (FileIO& outFile, double aSample);
void  soundWriteSample64F_L     (FileIO& outFile, double aSample);



//////////////////////////////
//
// SoundFileIn::determineWritingFunction --
//

void SoundFileWrite::determineWritingFunction(void) {
   switch (sndHeader.getOutputType()) {
      case TYPE_SND:                           // bigendian samples
         endianType = SIG_BIG_ENDIAN;
         switch (sndHeader.getNextFormat()) {
            case SND_FORMAT_MULAW_8:
               writeSample = soundWriteSample8M_B;
               break;
            case SND_FORMAT_LINEAR_8:
               writeSample = soundWriteSample8L_B;
               break;
            case SND_FORMAT_LINEAR_16:
               writeSample = soundWriteSample16L_B;
               break;
            case SND_FORMAT_LINEAR_24:
               writeSample = soundWriteSample24L_B;
               break;
            case SND_FORMAT_LINEAR_32:
               writeSample = soundWriteSample32L_B;
               break;
            case SND_FORMAT_FLOAT:
               writeSample = soundWriteSample32F_B;
               break;
            case SND_FORMAT_DOUBLE:
               writeSample = soundWriteSample64F_B;
               break;
            default:
               cerr << "Error: unknown output soundfile format: "
                    << sndHeader.getNextFormat() << endl;
               exit(1);
         }
         break;
      case TYPE_WAV_PCM:                      // little endian samples
         endianType = SIG_LITTLE_ENDIAN;
         switch (sndHeader.getNextFormat()) {
            case SND_FORMAT_MULAW_8:
               writeSample = soundWriteSample8M_L;
               break;
            case SND_FORMAT_LINEAR_8:
               writeSample = soundWriteSample8L_L;
               break;
            case SND_FORMAT_LINEAR_16:
               writeSample = soundWriteSample16L_L;
               break;
            case SND_FORMAT_LINEAR_24:
               writeSample = soundWriteSample24L_L;
               break;
            case SND_FORMAT_LINEAR_32:
               writeSample = soundWriteSample32L_L;
               break;
            case SND_FORMAT_FLOAT:
               writeSample = soundWriteSample32F_L;
               break;
            case SND_FORMAT_DOUBLE:
               writeSample = soundWriteSample64F_L;
               break;
            default:
               cerr << "Error: unknown output soundfile format: "
                    << sndHeader.getNextFormat() << endl;
               exit(1);
         }
         break;
      default:
         cerr << "Error: unknown soundfile type: "
              << sndHeader.getOutputType() << endl;
         exit(1);
   }
}



//////////////////////////////
//
// soundWriteSample8M -- read channel samples as
//      8-bit Mulaw values
//

void soundWriteSample8M_B(FileIO& outFile, double aSample) {
   outFile.writeBigEndian(linear16ToMulaw8(SampleToShort(aSample)));
}


void soundWriteSample8M_L(FileIO& outFile, double aSample) {
   outFile.writeLittleEndian(linear16ToMulaw8(SampleToShort(aSample)));
}



//////////////////////////////
//
// soundWriteSample8L
//

void soundWriteSample8L_B(FileIO& outFile, double aSample) {
   outFile.writeBigEndian(SampleToChar(aSample));
}


void soundWriteSample8L_L(FileIO& outFile, double aSample) {
   outFile.writeBigEndian(SampleToChar(aSample));
}



//////////////////////////////
//
// soundWriteSample16L
//

void soundWriteSample16L_B(FileIO& outFile, double aSample) {
   outFile.writeBigEndian(SampleToShort(aSample));
}


void soundWriteSample16L_L(FileIO& outFile, double aSample) {
   outFile.writeLittleEndian(SampleToShort(aSample));
}



//////////////////////////////
//
// soundWriteSample24L
//

void soundWriteSample24L_B(FileIO& outFile, double aSample) {
   static long output;
   output = SampleToLong(aSample);
   outFile.writeBigEndian((uchar)(output >> 12));
   outFile.writeBigEndian((uchar)(output >>  8));
   outFile.writeBigEndian((uchar)(output >>  4));
}


void soundWriteSample24L_L(FileIO& outFile, double aSample) {
   static long output;
   output = SampleToLong(aSample);
   outFile.writeBigEndian((uchar)(output >>  4));
   outFile.writeBigEndian((uchar)(output >>  8));
   outFile.writeBigEndian((uchar)(output >> 12));
}



//////////////////////////////
//
// soundWriteSample32L
//

void soundWriteSample32L_B(FileIO& outFile, double aSample) {
   outFile.writeBigEndian(SampleToLong(aSample));
}


void soundWriteSample32L_L(FileIO& outFile, double aSample) {
   outFile.writeBigEndian(SampleToLong(aSample));
}



//////////////////////////////
//
// soundWriteSample32F
//

void soundWriteSample32F_B(FileIO& outFile, double aSample) {
   outFile.writeBigEndian((float)aSample);
}


void soundWriteSample32F_L(FileIO& outFile, double aSample) {
   outFile.writeLittleEndian((float)aSample);
}



//////////////////////////////
//
// soundWriteSample64F
//

void soundWriteSample64F_B(FileIO& outFile, double aSample) {
   outFile.writeBigEndian((double)aSample);
}


void soundWriteSample64F_L(FileIO& outFile, double aSample) {
   outFile.writeLittleEndian((double)aSample);
}



//////////////////////////////
//
// linear16ToMulaw8
//
// reference: http://www.omnigroup.com/MailArchive/next-prog/1994/0042.html
//

#define BIAS 0x84 /* define the add-in bias for 16 bit samples */
#define CLIP 32635

uchar linear16ToMulaw8(short aSample) {

   static int exponentLookupTable[256] = {
      0,0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
      4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
      5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
      5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
      6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
      6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
      6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
      6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7
   };

   static int sign, exponent, mantissa, magnitude;
   static uchar output;

   // put the sample into sign-magnitude form
   sign = (aSample >> 8) & 0x80;                  // et aside the sign 
   magnitude = sign ? -aSample : aSample;         // get magnitude
   if (magnitude > CLIP) magnitude = CLIP;        // clip the magnitude 

   // Convert from 16 bit linear to ulaw. 
   magnitude += BIAS;
   exponent   = exponentLookupTable[(uchar)((magnitude >> 7) & 0xFF)];
   mantissa   = (magnitude >> (exponent + 3)) & 0x0F;
   output     = ~(sign | (exponent << 4) | mantissa);
   
   return output;
}



//////////////////////////////
//
// SampleToShort -- converts a number in the range -1.0 to 1.0
//	to a number in the range -32768 to 32767.  If the number
//      is out of the input range then it will be limited to the
//	maximum in the direction (and not wrap around).
//

short SampleToShort(double aSample) {
   int output;

     if (aSample > 0) {
        output = (int)(aSample * 0x8000 + 0.5);
     } else {
        output = (int)(aSample * 0x8000 - 0.5);
     }


   // prevent wrap around:
   if (output > 32767) output = 32767;
   else if (output < -32768) output = -32768;

   return (short)output;
}
   


//////////////////////////////
//
// SampleToLong -- converts a number in the range -1.0 to 1.0
//	to a number in the range -2147483647 to 2147483648.  If the 
//      number is out of the input range then it will be limited to 
//	the maximum in that direction (and not wrap around).
//

long SampleToLong(double aSample) {
   static long output;

   // prevent wrap around:
   if (aSample >= 1.0) {
      output = (long)0x80000000;
   } else if (aSample <= -1.0) {
      output = (long)0x7fffffff;
   } else {
      output = (long)(aSample * 0x7fffffff + 0.5);
   }

   return output;
}



//////////////////////////////
//
// SampleToChar -- converts a number in the range -1.0 to 1.0
//	to a number in the range -127 to 128.  If the 
//      number is out of the input range then it will be limited to 
//	the maximum in that direction (and not wrap around).
//

char SampleToChar(double aSample) {
   static char output;

   // prevent wrap around:
   if (aSample >= 1.0) {
      output = (char)0x7f;
   } else if (aSample <= -1.0) {
      output = (char)0x80;
   } else {
      output = (char)(aSample * 0x7f + 0.5);
   }

   return output;
}



